// gcc -o sq23video sq23video.c -lopencv_core -lopencv_imgproc -lopencv_highgui -lm
// ./sq23video udp://192.168.33.1:10900

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <time.h>

typedef int SOCKET;
#define SD_BOTH SHUT_RDWR
#define closesocket(sock) close((sock))
#endif // _WIN32

#include "opencv2/core/version.hpp"

// Temporary workaround because many conversions from C to C++ types are disabled by default after OpenCV 4.2.0...
#ifndef DISABLE_OPENCV_C_API_CTORS
#if (CV_MAJOR_VERSION >= 4)
#if (CV_MINOR_VERSION >= 2)
#ifndef CV__ENABLE_C_API_CTORS
#define CV__ENABLE_C_API_CTORS
#endif // !CV__ENABLE_C_API_CTORS
#ifndef CV__SKIP_MESSAGE_MALFORMED_C_API_CTORS
#define CV__SKIP_MESSAGE_MALFORMED_C_API_CTORS
#endif // !CV__SKIP_MESSAGE_MALFORMED_C_API_CTORS
#endif // (CV_MINOR_VERSION >= 2)
#endif // (CV_MAJOR_VERSION >= 4)
#endif // !DISABLE_OPENCV_C_API_CTORS

#include "opencv2/core/core_c.h"
#include "opencv2/imgproc/imgproc_c.h"
#include "opencv2/highgui/highgui_c.h"
#if (CV_MAJOR_VERSION == 3)
#include "opencv2/imgcodecs/imgcodecs_c.h"
#include "opencv2/videoio/videoio_c.h"
#endif // (CV_MAJOR_VERSION == 3)
#if (CV_MAJOR_VERSION >= 4)
#include "opencv2/imgcodecs/legacy/constants_c.h"
#include "opencv2/videoio/legacy/constants_c.h"
#include "opencv2/videoio/videoio_c.h"
#endif // (CV_MAJOR_VERSION >= 4)

#include "opencv2/core/core.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#if (CV_MAJOR_VERSION >= 3)
#include "opencv2/imgcodecs/imgcodecs.hpp"
#include "opencv2/videoio/videoio.hpp"
#endif // (CV_MAJOR_VERSION >= 3)

#define UDP_PACKET_SIZE 1450

int main(int argc, char *argv[])
{
	SOCKET s;
	struct sockaddr_in sai;
	struct sockaddr_storage sa;
	socklen_t salen = 0;
	char buf[UDP_PACKET_SIZE];
	int ret = 0;
	unsigned int filenumberrstartingfrom1 = 0;
	unsigned int nbofpackets = 0;
	unsigned int packetnumberstartingfrom0 = 0;
	char* imagebuf = (char*)calloc(256, UDP_PACKET_SIZE);
	unsigned int imagebuflen = 0;
	unsigned int curfilenumberrstartingfrom1 = 0;
	unsigned int curnbofpacketsrecv = 0;
	IplImage* iplimg = NULL;

#ifdef _WIN32
	WSADATA wsaData;

	WSAStartup(MAKEWORD(2, 2), &wsaData);
#endif // _WIN32

	s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	memset(&sai, 0, sizeof(sai));
	sai.sin_family = AF_INET;
	sai.sin_port = htons((unsigned short)atoi("10900"));
	sai.sin_addr.s_addr = inet_addr("0.0.0.0");
	ret = bind(s, (struct sockaddr*)&sai, sizeof(sai));

	cvNamedWindow("cam", CV_WINDOW_AUTOSIZE);
	iplimg = cvCreateImage(cvSize(640, 384), IPL_DEPTH_8U, 3);

	for (;;)
	{
		memset(&sa, 0, sizeof(sa));
		salen = sizeof(sa);
		memset(buf, 0, sizeof(buf));
		ret = recvfrom(s, buf, sizeof(buf), 0, (struct sockaddr*)&sa, (socklen_t*)&salen);
		if (ret != -1)
		{
			filenumberrstartingfrom1 = (buf[1]<<8)+buf[0];
			nbofpackets = buf[2];
			packetnumberstartingfrom0 = buf[3];
			//printf("%u %u %u\n", filenumberrstartingfrom1, nbofpackets, packetnumberstartingfrom0);

			if (packetnumberstartingfrom0 >= nbofpackets) 
			{ 
				printf("Please restart the camera\n");
				continue; 
			}

			if ((packetnumberstartingfrom0 == 0)&&(curnbofpacketsrecv == 0)) { curfilenumberrstartingfrom1 = filenumberrstartingfrom1; curnbofpacketsrecv = 0;}

			if (curfilenumberrstartingfrom1 == filenumberrstartingfrom1)
			{
				//curnbofpacketsrecv++;

				// 0xFFD8 means SOI(Start of image), 0xFFD9 means EOI(End of image), see https://www.media.mit.edu/pia/Research/deepview/exif.html#:~:text=JPEG%20image%20files.-,JPEG%20format%20and%20Marker,EOI(End%20of%20image).
				int offset = 8;
				if (packetnumberstartingfrom0 == 0) memcpy(imagebuf, buf+24, ret-24);
				else memcpy(imagebuf+UDP_PACKET_SIZE-24+(packetnumberstartingfrom0-1)*(UDP_PACKET_SIZE-offset), buf+offset, ret-offset);

				if (packetnumberstartingfrom0 == nbofpackets-1)//&&(curnbofpacketsrecv == nbofpackets))
				{
					curnbofpacketsrecv = 0;
					//curfilenumberrstartingfrom1 = filenumberrstartingfrom1+1;

					imagebuflen = UDP_PACKET_SIZE-24+(nbofpackets-2)*(UDP_PACKET_SIZE-offset)+ret-offset;

					CvMat* mat = NULL;

					mat = cvCreateMat(1, imagebuflen, CV_8UC1);
					if (mat == NULL)
					{
						printf("cvCreateMat() failed.\n");
						continue;
					}

					memcpy(mat->data.ptr, imagebuf, imagebuflen);

					cv::Mat _matnew = cv::imdecode(cv::cvarrToMat(mat), cv::IMREAD_COLOR);
					IplImage _imagenew = _matnew;
					iplimg = &_imagenew;
					if (iplimg == NULL)
					{
						printf("cvDecodeImage() failed.\n");
						cvReleaseMat(&mat);
						continue;
					}

					cvShowImage("cam", iplimg);
					cvWaitKey(10);
	
					cvReleaseMat(&mat);
				}		
			}
		}
	}

	cvReleaseImage(&iplimg);
	cvDestroyWindow("cam");

	closesocket(s);

#ifdef _WIN32
	WSACleanup();
#endif // _WIN32

	free(imagebuf);

	return 0;
}
